class FireboltError(Exception):
    """Base class for all Firebolt errors."""


class FireboltEngineError(FireboltError):
    """Base class for engine related errors."""


class EngineNotRunningError(FireboltEngineError):
    """Engine, that's being queried, is not running."""


class NoAttachedDatabaseError(FireboltEngineError):
    """Engine, that's being accessed, is not running.

    Args:
        method_name (str): Name of the method, which caused issue

    Attributes:
        method_name (str): Name of the method, which caused issue
    """

    def __init__(self, method_name: str):
        super().__init__(
            f"Unable to call {method_name}: "
            "Engine must to be attached to a database first."
        )
        self.method_name = method_name


class AlreadyBoundError(FireboltEngineError):
    """Engine is already bound to a database."""


class FireboltDatabaseError(FireboltError):
    """Base class for database related errors."""


class AccountNotFoundError(FireboltError):
    """Account with provided name doesn't exist.

    Args:
        account_name (str): Name of account, that wasn't found

    Attributes:
        account_name (str): Name of account, that wasn't found
    """

    def __init__(self, account_name: str):
        super().__init__(f"Account '{account_name}' does not exist.")
        self.account_name = account_name


class AttachedEngineInUseError(FireboltDatabaseError):
    """Engine is unavailable because it's starting/stopping.

    Args:
        method_name (str): Name of the method, which caused issue

    Attributes:
        method_name (str): Name of the method, which caused issue
    """

    def __init__(self, method_name: str):
        super().__init__(
            f"Unable to call {method_name}: "
            "Engine must not be in starting or stopping state."
        )
        self.method_name = method_name


class ConnectionError(FireboltError):
    """Base class for connection related errors."""


class ConnectionClosedError(ConnectionError):
    """Connection operations are unavailable since it's closed."""


class CursorError(FireboltError):
    """Base class for cursor related errors."""


class CursorClosedError(CursorError):
    """Cursor operations are unavailable since it's closed.

    Args:
        method_name (str): Name of the method, which caused issue

    Attributes:
        method_name (str): Name of the method, which caused issue
    """

    def __init__(self, method_name: str):
        super().__init__(f"Unable to call {method_name}: cursor closed.")
        self.method_name = method_name


class QueryNotRunError(CursorError):
    """Some cursor methods are unavailable before a query is run.

    Args:
        method_name (str): Name of the method, which caused issue

    Attributes:
        method_name (str): Name of the method, which caused issue
    """

    def __init__(self, method_name: str):
        super().__init__(f"Unable to call {method_name}: need to run a query first.")
        self.method_name = method_name


class AuthenticationError(FireboltError):
    """Firebolt authentication error.

    Stores error cause and authentication endpoint.

    Args:
        api_endpoint (str): Environment api endpoint, used for authentication
        cause (str): Reason for authentication failure

    Attributes:
        api_endpoint (str): Environment api endpoint, used for authentication
        cause (str): Reason for authentication failure
    """

    def __init__(self, cause: str):
        super().__init__(f"Failed to authenticate: {cause}.")
        self.cause = cause


class AuthorizationError(FireboltError):
    """Firebolt authentication error.

    Args:
        cause (str): Reason for authorization failure

    Attributes:
        cause (str): Reason for authorization failure
    """

    def __init__(self, cause: str):
        super().__init__(f"Authorization failed: {cause}.")


# PEP-249


class Warning(Exception):
    """Base class for warning exceptions.

    Exception raised for important warnings, like data truncations while inserting, etc.
    """


Error = FireboltError


class InterfaceError(Error):
    """Database interface related error.

    Exception raised for errors that are related to the database interface
    rather than the database itself.
    """


class DatabaseError(Error):
    """Exception raised for errors that are related to the database."""


class DataError(DatabaseError):
    """Data processing error.

    Exception raised for errors that are due to problems with the processed data
    like division by zero, numeric value out of range, etc.
    """


class OperationalError(DatabaseError):
    """Database operating error.

    Exception raised for errors that are related to the database's operation
    and not necessarily under the control of the programmer, e.g. an unexpected
    disconnect occurs, the data source name is not found, a transaction could not
    be processed, a memory allocation error occurred during processing, etc.
    """


class IntegrityError(DatabaseError):
    """Database data integrity error.

    Exception raised when the relational integrity of the database is affected,
    e.g. a foreign key check fails.
    """


class InternalError(DatabaseError):
    """Database internal error.

    Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc.
    """


class ProgrammingError(DatabaseError):
    """Database programming error.

    Exception raised for programming errors, e.g. table not found or already exists,
    syntax error in the SQL statement, wrong number of parameters specified, etc.
    """


class NotSupportedError(DatabaseError):
    """Operation not supported.

    Exception raised in case a method or database API was used which is not supported
    by the database, e.g. requesting a .rollback() on a connection that
    does not support transaction or has transactions turned off.
    """


class ConfigurationError(InterfaceError):
    """Invalid configuration error."""



from firebolt.client.auth import Auth  # backward compatibility
from firebolt.client.client import AsyncClient, Client
from firebolt.client.constants import DEFAULT_API_URL
from firebolt.client.resource_manager_hooks import (
    log_request,
    log_response,
    raise_on_4xx_5xx,
)



from typing import Generator

from httpx import Request, Response

from firebolt.client.auth import Auth
from firebolt.utils.exception import AuthorizationError


class Token(Auth):
    """Token authentication class for Firebolt database.

    Uses provided token for authentication. Doesn't cache token and doesn't
    refresh it on expiration

    Args:
        token (str): Authorization token

    Attributes:
        token (str):
    """

    def __init__(self, token: str):
        super().__init__(use_token_cache=False)
        self._token = token

    def copy(self) -> "Token":
        """Make another auth object with same credentials.

        Returns:
            Token: Auth object
        """
        assert self.token
        return Token(self.token)

    def get_new_token_generator(self) -> Generator[Request, Response, None]:
        """Raise authorization error since token is invalid or expired.

        Raises:
            AuthorizationError: Token is invalid or expired
        """
        raise AuthorizationError("Provided token in not valid anymore.")



from time import time
from typing import Generator, Optional

from httpx import Request, Response

from firebolt.client.auth.base import Auth
from firebolt.client.constants import _REQUEST_ERRORS
from firebolt.utils.exception import AuthenticationError
from firebolt.utils.token_storage import TokenSecureStorage
from firebolt.utils.urls import AUTH_URL
from firebolt.utils.util import cached_property


class UsernamePassword(Auth):
    """Username/Password authentication class for Firebolt database.

    Gets authentication token using
    provided credentials and updates it when it expires

    Args:
        username (str): Username
        password (str): Password
        use_token_cache (bool): True if token should be cached in filesystem,
            False otherwise

    Attributes:
        username (str): Username
        password (str): Password
    """

    __slots__ = (
        "username",
        "password",
        "_token",
        "_expires",
        "_use_token_cache",
    )

    requires_response_body = True

    def __init__(
        self,
        username: str,
        password: str,
        use_token_cache: bool = True,
    ):
        self.username = username
        self.password = password
        super().__init__(use_token_cache)

    def copy(self) -> "UsernamePassword":
        """Make another auth object with same credentials.

        Returns:
            UsernamePassword: Auth object
        """
        return UsernamePassword(self.username, self.password, self._use_token_cache)

    @cached_property
    def _token_storage(self) -> Optional[TokenSecureStorage]:
        """Token filesystem cache storage.

        This is evaluated lazily, only if caching is enabled

        Returns:
            TokenSecureStorage: Token filesystem cache storage
        """
        return TokenSecureStorage(username=self.username, password=self.password)

    def get_new_token_generator(self) -> Generator[Request, Response, None]:
        """Get new token using username and password.

        Yields:
            Request: An http request to get token. Expects Response to be sent back

        Raises:
            AuthenticationError: Error while authenticating with provided credentials
        """
        try:
            response = yield self.request_class(
                "POST",
                # The full url is generated on client side by attaching
                # it to api_endpoint
                AUTH_URL,
                headers={
                    "Content-Type": "application/json;charset=UTF-8",
                    "User-Agent": "firebolt-sdk",
                },
                json={"username": self.username, "password": self.password},
            )
            response.raise_for_status()

            parsed = response.json()
            self._check_response_error(parsed)

            self._token = parsed["access_token"]
            self._expires = int(time()) + int(parsed["expires_in"])

        except _REQUEST_ERRORS as e:
            raise AuthenticationError(repr(e)) from e

    def _check_response_error(self, response: dict) -> None:
        """Check if response data contains errors.

        Args:
            response (dict): Response data

        Raises:
            AuthenticationError: Were unable to authenticate
        """
        if "error" in response:
            raise AuthenticationError(
                response.get("message", "unknown server error"),
            )



from logging import getLogger
from typing import Any, Dict, List, Optional, Union

from firebolt.model.engine import Engine, EngineSettings, _EngineCreateRequest
from firebolt.model.engine_revision import (
    EngineRevision,
    EngineRevisionSpecification,
)
from firebolt.model.region import Region
from firebolt.service.base import BaseService
from firebolt.service.types import EngineOrder, EngineType, WarmupMethod
from firebolt.utils.exception import FireboltError
from firebolt.utils.urls import (
    ACCOUNT_ENGINE_BY_NAME_URL,
    ACCOUNT_ENGINE_URL,
    ACCOUNT_ENGINES_URL,
    ENGINES_BY_IDS_URL,
)
from firebolt.utils.util import prune_dict

logger = getLogger(__name__)


class EngineService(BaseService):
    def get(self, id_: str) -> Engine:
        """Get an Engine from Firebolt by its id."""

        response = self.client.get(
            url=ACCOUNT_ENGINE_URL.format(account_id=self.account_id, engine_id=id_),
        )
        engine_entry: dict = response.json()["engine"]
        return Engine.parse_obj_with_service(obj=engine_entry, engine_service=self)

    def get_by_ids(self, ids: List[str]) -> List[Engine]:
        """Get multiple Engines from Firebolt by their ids."""
        response = self.client.post(
            url=ENGINES_BY_IDS_URL,
            json={
                "engine_ids": [
                    {"account_id": self.account_id, "engine_id": engine_id}
                    for engine_id in ids
                ]
            },
        )
        return [
            Engine.parse_obj_with_service(obj=e, engine_service=self)
            for e in response.json()["engines"]
        ]

    def get_by_name(self, name: str) -> Engine:
        """Get an Engine from Firebolt by its name."""

        response = self.client.get(
            url=ACCOUNT_ENGINE_BY_NAME_URL.format(account_id=self.account_id),
            params={"engine_name": name},
        )
        engine_id = response.json()["engine_id"]["engine_id"]
        return self.get(id_=engine_id)

    def get_many(
        self,
        name_contains: Optional[str] = None,
        current_status_eq: Optional[str] = None,
        current_status_not_eq: Optional[str] = None,
        region_eq: Optional[str] = None,
        order_by: Optional[Union[str, EngineOrder]] = None,
    ) -> List[Engine]:
        """
        Get a list of engines on Firebolt.

        Args:
            name_contains: Filter for engines with a name containing this substring.
            current_status_eq: Filter for engines with this status.
            current_status_not_eq: Filter for engines that do not have this status.
            region_eq: Filter for engines by region.
            order_by: Method by which to order the results. See [EngineOrder].

        Returns:
            A list of engines matching the filters.
        """

        if isinstance(order_by, str):
            order_by = EngineOrder[order_by].name

        if region_eq is not None:
            region_eq = self.resource_manager.regions.get_by_name(
                name=region_eq
            ).key.region_id

        response = self.client.get(
            url=ACCOUNT_ENGINES_URL.format(account_id=self.account_id),
            params=prune_dict(
                {
                    "page.first": 5000,  # FUTURE: pagination support w/ generator
                    "filter.name_contains": name_contains,
                    "filter.current_status_eq": current_status_eq,
                    "filter.current_status_not_eq": current_status_not_eq,
                    "filter.compute_region_id_region_id_eq": region_eq,
                    "order_by": order_by,
                }
            ),
        )
        return [
            Engine.parse_obj_with_service(obj=e["node"], engine_service=self)
            for e in response.json()["edges"]
        ]

    def create(
        self,
        name: str,
        region: Union[str, Region, None] = None,
        engine_type: Union[str, EngineType] = EngineType.GENERAL_PURPOSE,
        scale: int = 2,
        spec: Optional[str] = None,
        auto_stop: int = 20,
        warmup: Union[str, WarmupMethod] = WarmupMethod.PRELOAD_INDEXES,
        description: str = "",
        engine_settings_kwargs: Dict[str, Any] = {},
        revision_spec_kwargs: Dict[str, Any] = {},
    ) -> Engine:
        """
        Create a new Engine.

        Args:
            name: An identifier that specifies the name of the engine.
            region: The AWS region in which the engine runs.
            engine_type: The engine type. GENERAL_PURPOSE or DATA_ANALYTICS
            scale: The number of compute instances on the engine.
                The scale can be any int from 1 to 128.
            spec: Firebolt instance type. If not set will default to
                the cheapest instance.
            auto_stop: The amount of time (in minutes)
            after which the engine automatically stops.
            warmup: The warmup method that should be used.

                `MINIMAL` - On-demand loading (both indexes and tables' data).

                `PRELOAD_INDEXES` - Load indexes only.

                `PRELOAD_ALL_DATA` - Full data auto-load
                (both indexes and table data - full warmup).
            description: A short description of the engine's purpose.

        Returns:
            Engine with the specified settings.
        """

        logger.info(f"Creating Engine (name={name})")

        if isinstance(engine_type, str):
            engine_type = EngineType[engine_type]
        if isinstance(warmup, str):
            warmup = WarmupMethod[warmup]

        if region is None:
            region = self.resource_manager.regions.default_region
        else:
            if isinstance(region, str):
                region = self.resource_manager.regions.get_by_name(name=region)

        engine = Engine(
            name=name,
            description=description,
            compute_region_key=region.key,
            settings=EngineSettings.default(
                engine_type=engine_type,
                auto_stop_delay_duration=f"{auto_stop * 60}s",
                warm_up=warmup,
                **engine_settings_kwargs,
            ),
        )

        if spec:
            instance_type_key = self.resource_manager.instance_types.get_by_name(
                instance_type_name=spec, region_name=region.name
            ).key
        else:
            instance_type = (
                self.resource_manager.instance_types.cheapest_instance_in_region(region)
            )
            if not instance_type:
                raise FireboltError(
                    f"No suitable default instances found in region {region}"
                )
            instance_type_key = instance_type.key

        engine_revision = EngineRevision(
            specification=EngineRevisionSpecification(
                db_compute_instances_type_key=instance_type_key,
                db_compute_instances_count=scale,
                proxy_instances_type_key=instance_type_key,
                **revision_spec_kwargs,
            )
        )

        return self._send_create_engine(engine=engine, engine_revision=engine_revision)

    def _send_create_engine(
        self, engine: Engine, engine_revision: Optional[EngineRevision] = None
    ) -> Engine:
        """
        Create a new Engine on Firebolt from the local Engine object.

        Args:
            engine: The Engine to create.
            engine_revision: EngineRevision to use for configuring the Engine.

        Returns:
            The newly created engine.
        """

        response = self.client.post(
            url=ACCOUNT_ENGINES_URL.format(account_id=self.account_id),
            headers={"Content-type": "application/json"},
            json=_EngineCreateRequest(
                account_id=self.account_id,
                engine=engine,
                engine_revision=engine_revision,
            ).jsonable_dict(by_alias=True),
        )
        return Engine.parse_obj_with_service(
            obj=response.json()["engine"], engine_service=self
        )



from asyncio import (
    AbstractEventLoop,
    get_event_loop,
    new_event_loop,
    set_event_loop,
)
from functools import lru_cache, wraps
from threading import Thread
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Optional,
    Type,
    TypeVar,
)

from httpx import URL

T = TypeVar("T")


def cached_property(func: Callable[..., T]) -> T:
    """cached_property implementation for 3.7 backward compatibility.

    Args:
        func (Callable): Property getter

    Returns:
        T: Property of type, returned by getter
    """
    return property(lru_cache()(func))  # type: ignore


def prune_dict(d: dict) -> dict:
    """Prune items from dictionaries where value is None.

    Args:
        d (dict): Dict to prune

    Returns:
        dict: Pruned dict
    """
    return {k: v for k, v in d.items() if v is not None}


TMix = TypeVar("TMix")


def mixin_for(baseclass: Type[TMix]) -> Type[TMix]:
    """Define mixin with baseclass typehint.

    Should be used as a mixin base class to fix typehints

    Args:
        baseclass (Type[TMix]): Class which mixin will be made for

    Returns:
        Type[TMix]: Mixin type to inherin from

    Examples:
        ```
        class ReadonlyMixin(mixin_for(BaseClass))):
            ...
        ```

    """
    if TYPE_CHECKING:
        return baseclass
    return object


def fix_url_schema(url: str) -> str:
    """Add schema to URL if it's missing.

    Args:
        url (str): URL to check

    Returns:
        str: URL with schema present

    """
    return url if url.startswith("http") else f"https://{url}"


class AsyncJobThread:
    """Thread runner that allows running async tasks syncronously in a separate thread.

    Caches loop to be reused in all threads.
    It allows running async functions syncronously inside a running event loop.
    Since nesting loops is not allowed, we create a separate thread for a new event loop

    Attributes:
        result (Any): Value, returned by coroutine execution
        exception (Optional[BaseException]): If any, exception that occured
            during coroutine execution
    """

    def __init__(self) -> None:
        self._loop: Optional[AbstractEventLoop] = None
        self.result: Any = None
        self.exception: Optional[BaseException] = None

    def _initialize_loop(self) -> None:
        """Initialize a loop once to use for later execution.

        Tries to get a running loop.
        Creates a new loop if no active one and sets it as active.
        """
        if not self._loop:
            try:
                # despite the docs, this function fails if no loop is set
                self._loop = get_event_loop()
            except RuntimeError:
                self._loop = new_event_loop()
        set_event_loop(self._loop)

    def _run(self, coro: Coroutine) -> None:
        """Run coroutine in an event loop.

        Execution return value is stored into ``result`` field
        If an exception occures, it will be caught and stored into ``exception`` field

        Args:
            coro (Coroutine): Coroutine to execute
        """
        try:
            self._initialize_loop()
            assert self._loop is not None
            self.result = self._loop.run_until_complete(coro)
        except BaseException as e:
            self.exception = e

    def execute(self, coro: Coroutine) -> Any:
        """Execute coroutine in a separate thread.

        Args:
            coro (Coroutine): Coroutine to execute

        Returns:
            Any: Coroutine execution return value

        Raises:
            exception: Exeption, occured within coroutine
        """
        thread = Thread(target=self._run, args=[coro])
        thread.start()
        thread.join()
        if self.exception:
            raise self.exception
        return self.result


def async_to_sync(f: Callable, async_job_thread: AsyncJobThread = None) -> Callable:
    """Convert async function to sync.

    Args:
        f (Callable): function to convert
        async_job_thread (AsyncJobThread): Job thread instance to use for async excution
            (Default value = None)

    Returns:
        Callable: regular function, which can be executed syncronously
    """

    @wraps(f)
    def sync(*args: Any, **kwargs: Any) -> Any:
        try:
            loop = get_event_loop()
        except RuntimeError:
            loop = new_event_loop()
            set_event_loop(loop)
        # We are inside a running loop
        if loop.is_running():
            nonlocal async_job_thread
            if not async_job_thread:
                async_job_thread = AsyncJobThread()
            return async_job_thread.execute(f(*args, **kwargs))
        return loop.run_until_complete(f(*args, **kwargs))

    return sync


def merge_urls(base: URL, merge: URL) -> URL:
    """Merge a base and merge urls.

    If merge is not a relative url, do nothing

    Args:
        base (URL): Base URL to merge to
        merge (URL): URL to merge

    Returns:
        URL: Resulting URL
    """
    if merge.is_relative_url:
        merge_raw_path = base.raw_path + merge.raw_path.lstrip(b"/")
        return base.copy_with(raw_path=merge_raw_path)
    return merge



from __future__ import annotations

import functools
import logging
import time
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence

from pydantic import Field, PrivateAttr

from firebolt.db import Connection, connect
from firebolt.model import FireboltBaseModel
from firebolt.model.binding import Binding
from firebolt.model.database import Database
from firebolt.model.engine_revision import EngineRevision, EngineRevisionKey
from firebolt.model.region import RegionKey
from firebolt.service.types import (
    EngineStatus,
    EngineStatusSummary,
    EngineType,
    WarmupMethod,
)
from firebolt.utils.exception import NoAttachedDatabaseError
from firebolt.utils.urls import (
    ACCOUNT_ENGINE_RESTART_URL,
    ACCOUNT_ENGINE_START_URL,
    ACCOUNT_ENGINE_STOP_URL,
    ACCOUNT_ENGINE_URL,
)
from firebolt.utils.util import prune_dict

if TYPE_CHECKING:
    from firebolt.service.engine import EngineService

logger = logging.getLogger(__name__)


class EngineKey(FireboltBaseModel):
    account_id: str
    engine_id: str


def wait(seconds: int, timeout_time: float, error_message: str, verbose: bool) -> None:
    time.sleep(seconds)
    if time.time() > timeout_time:
        raise TimeoutError(error_message)
    if verbose:
        print(".", end="")


class EngineSettings(FireboltBaseModel):
    """
    Engine Settings.

    See Also: :py:class:`EngineRevisionSpecification
    <firebolt.model.engine_revision.EngineRevisionSpecification>`
    which also contains engine configuration.
    """

    preset: str
    auto_stop_delay_duration: str = Field(regex=r"^[0-9]+[sm]$|^0$")
    minimum_logging_level: str
    is_read_only: bool
    warm_up: str

    @classmethod
    def default(
        cls,
        engine_type: EngineType = EngineType.GENERAL_PURPOSE,
        auto_stop_delay_duration: str = "1200s",
        warm_up: WarmupMethod = WarmupMethod.PRELOAD_INDEXES,
        minimum_logging_level: str = "ENGINE_SETTINGS_LOGGING_LEVEL_INFO",
    ) -> EngineSettings:
        if engine_type == EngineType.GENERAL_PURPOSE:
            preset = engine_type.GENERAL_PURPOSE.api_settings_preset_name  # type: ignore # noqa: E501
            is_read_only = False
        else:
            preset = engine_type.DATA_ANALYTICS.api_settings_preset_name  # type: ignore
            is_read_only = True

        return cls(
            preset=preset,
            auto_stop_delay_duration=auto_stop_delay_duration,
            minimum_logging_level=minimum_logging_level,
            is_read_only=is_read_only,
            warm_up=warm_up.api_name,
        )


def check_attached_to_database(func: Callable) -> Callable:
    """(Decorator) Ensure the engine is attached to a database."""

    @functools.wraps(func)
    def inner(self: Engine, *args: Any, **kwargs: Any) -> Any:
        if self.database is None:
            raise NoAttachedDatabaseError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


class FieldMask(FireboltBaseModel):
    paths: Sequence[str] = Field(alias="paths")


class Engine(FireboltBaseModel):
    """
    A Firebolt engine. Responsible for performing work (queries, data ingestion).

    Engines are configured in :py:class:`Settings
    <firebolt.model.engine.EngineSettings>`
    and in :py:class:`EngineRevisionSpecification
    <firebolt.model.engine_revision.EngineRevisionSpecification>`.
    """

    # internal
    _service: EngineService = PrivateAttr()

    # required
    name: str = Field(min_length=1, max_length=255, regex=r"^[0-9a-zA-Z_]+$")
    compute_region_key: RegionKey = Field(alias="compute_region_id")
    settings: EngineSettings

    # optional
    key: Optional[EngineKey] = Field(alias="id")
    description: Optional[str]
    emoji: Optional[str]
    current_status: Optional[EngineStatus]
    current_status_summary: Optional[EngineStatusSummary]
    latest_revision_key: Optional[EngineRevisionKey] = Field(alias="latest_revision_id")
    endpoint: Optional[str]
    endpoint_serving_revision_key: Optional[EngineRevisionKey] = Field(
        alias="endpoint_serving_revision_id"
    )
    create_time: Optional[datetime]
    create_actor: Optional[str]
    last_update_time: Optional[datetime]
    last_update_actor: Optional[str]
    last_use_time: Optional[datetime]
    desired_status: Optional[str]
    health_status: Optional[str]
    endpoint_desired_revision_key: Optional[EngineRevisionKey] = Field(
        alias="endpoint_desired_revision_id"
    )

    @classmethod
    def parse_obj_with_service(cls, obj: Any, engine_service: EngineService) -> Engine:
        engine = cls.parse_obj(obj)
        engine._service = engine_service
        return engine

    @property
    def engine_id(self) -> str:
        if self.key is None:
            raise ValueError("engine key is None")
        return self.key.engine_id

    @property
    def database(self) -> Optional[Database]:
        return self._service.resource_manager.bindings.get_database_bound_to_engine(
            engine=self
        )

    def get_latest(self) -> Engine:
        """Get an up-to-date instance of the Engine from Firebolt."""
        return self._service.get(id_=self.engine_id)

    def attach_to_database(
        self, database: Database, is_default_engine: bool = False
    ) -> Binding:
        """
        Attach this engine to a database.

        Args:
            database: Database to which the engine will be attached.
            is_default_engine:
                Whether this engine should be used as default for this database.
                Only one engine can be set as default for a single database.
                This will overwrite any existing default.
        """
        return self._service.resource_manager.bindings.create(
            engine=self, database=database, is_default_engine=is_default_engine
        )

    @check_attached_to_database
    def get_connection(self) -> Connection:
        """Get a connection to the attached database, for running queries.

        Returns:
            Connection: engine connection instance

        """
        return connect(
            database=self.database.name,  # type: ignore # already checked by decorator
            auth=self._service.client.auth,
            engine_url=self.endpoint,
            account_name=self._service.settings.account_name,
            api_endpoint=self._service.settings.server,
        )

    @check_attached_to_database
    def start(
        self,
        wait_for_startup: bool = True,
        wait_timeout_seconds: int = 3600,
        verbose: bool = False,
    ) -> Engine:
        """
        Start an engine. If it's already started, do nothing.

        Args:
            wait_for_startup:
                If True, wait for startup to complete.
                If false, return immediately after requesting startup.
            wait_timeout_seconds:
                Number of seconds to wait for startup to complete
                before raising a TimeoutError.
            verbose:
                If True, print dots periodically while waiting for engine startup.
                If false, do not print any dots.

        Returns:
            The updated Engine from Firebolt.
        """
        timeout_time = time.time() + wait_timeout_seconds

        engine = self.get_latest()
        if (
            engine.current_status_summary
            == EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING
        ):
            logger.info(
                f"Engine (engine_id={self.engine_id}, name={self.name}) "
                "is already running."
            )
            return engine

        # wait for engine to stop first, if it's already stopping
        # FUTURE: revisit logging and consider consolidating this if & the while below.
        elif (
            engine.current_status_summary
            == EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPING
        ):
            logger.info(
                f"Engine (engine_id={engine.engine_id}, name={engine.name}) "
                "is in currently stopping, waiting for it to stop first."
            )
            while (
                engine.current_status_summary
                != EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED
            ):
                wait(
                    seconds=5,
                    timeout_time=timeout_time,
                    error_message=(
                        "Engine "
                        f"(engine_id={engine.engine_id}, name={engine.name}) "
                        f"did not stop within {wait_timeout_seconds} seconds."
                    ),
                    verbose=True,
                )
                engine = engine.get_latest()

            logger.info(
                f"Engine (engine_id={engine.engine_id}, name={engine.name}) stopped."
            )

        engine = self._send_engine_request(ACCOUNT_ENGINE_START_URL)
        logger.info(
            f"Starting Engine (engine_id={engine.engine_id}, name={engine.name})"
        )

        # wait for engine to start
        while wait_for_startup and engine.current_status_summary not in {
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED,
        }:
            wait(
                seconds=5,
                timeout_time=timeout_time,
                error_message=(  # noqa: E501
                    f"Could not start engine within {wait_timeout_seconds} seconds."
                ),
                verbose=verbose,
            )
            previous_status_summary = engine.current_status_summary
            engine = engine.get_latest()
            if engine.current_status_summary != previous_status_summary:
                logger.info(
                    "Engine status_summary="
                    f"{getattr(engine.current_status_summary, 'name')}"
                )

        return engine

    @check_attached_to_database
    def stop(
        self, wait_for_stop: bool = False, wait_timeout_seconds: int = 3600
    ) -> Engine:
        """Stop an Engine running on Firebolt."""
        timeout_time = time.time() + wait_timeout_seconds

        engine = self._send_engine_request(ACCOUNT_ENGINE_STOP_URL)
        logger.info(f"Stopping Engine (engine_id={self.engine_id}, name={self.name})")

        while wait_for_stop and engine.current_status_summary not in {
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED,
        }:
            wait(
                seconds=5,
                timeout_time=timeout_time,
                error_message=(  # noqa: E501
                    f"Could not stop engine within {wait_timeout_seconds} seconds."
                ),
                verbose=False,
            )

            engine = engine.get_latest()

        return engine

    def update(
        self,
        name: Optional[str] = None,
        engine_type: Optional[EngineType] = None,
        scale: Optional[int] = None,
        spec: Optional[str] = None,
        auto_stop: Optional[int] = None,
        warmup: Optional[WarmupMethod] = None,
        description: Optional[str] = None,
        use_spot: Optional[bool] = None,
    ) -> Engine:
        """
        update the engine, and returns an updated version of the engine, all parameters
        could be set to None, this would keep the old engine parameter value
        """

        class _EngineUpdateRequest(FireboltBaseModel):
            """Helper model for sending Engine update requests."""

            account_id: str
            desired_revision: Optional[EngineRevision]
            engine: Engine
            engine_id: str
            update_mask: FieldMask

        # Update the engine parameters
        self.name = name if name else self.name
        self.description = description

        # Update engine settings
        engine_settings_params = {
            "engine_type": engine_type,
            "auto_stop_delay_duration": f"{auto_stop * 60}s"
            if auto_stop is not None
            else None,
            "warm_up": warmup,
        }
        self.settings = EngineSettings.default(**prune_dict(engine_settings_params))

        # Update the engine desired_revision if needed
        desired_revision = None
        if (scale or spec or use_spot is not None) and self.latest_revision_key:
            rm = self._service.resource_manager
            desired_revision = rm.engine_revisions.get_by_key(self.latest_revision_key)

            if spec:
                instance_type_key = rm.instance_types.get_by_name(
                    instance_type_name=spec,
                    region_name=rm.regions.regions_by_key[self.compute_region_key].name,
                ).key

                desired_revision.specification.db_compute_instances_type_key = (
                    instance_type_key
                )
                desired_revision.specification.proxy_instances_type_key = (
                    instance_type_key
                )

            if scale:
                desired_revision.specification.db_compute_instances_count = scale

            if use_spot is not None:
                desired_revision.specification.db_compute_instances_use_spot = use_spot

        update_mask_paths = list(
            prune_dict(
                {
                    "name": name,
                    "description": description,
                    "settings.auto_stop_delay_duration": auto_stop,
                    "settings.warm_up": warmup,
                    "settings.is_read_only": engine_type,
                    "settings.preset": engine_type,
                }
            ).keys()
        )

        # Send the update request
        response = self._service.client.patch(
            url=ACCOUNT_ENGINE_URL.format(
                engine_id=self.engine_id, account_id=self._service.account_id
            ),
            headers={"Content-type": "application/json"},
            json=_EngineUpdateRequest(
                account_id=self._service.account_id,
                engine=self,
                desired_revision=desired_revision,
                engine_id=self.engine_id,
                update_mask=FieldMask(paths=update_mask_paths),
            ).jsonable_dict(by_alias=True),
        )

        return Engine.parse_obj_with_service(
            obj=response.json()["engine"], engine_service=self._service
        )

    @check_attached_to_database
    def restart(
        self,
        wait_for_startup: bool = True,
        wait_timeout_seconds: int = 3600,
    ) -> Engine:
        """
        Restart an engine.

        Args:
            wait_for_startup:
                If True, wait for startup to complete.
                If false, return immediately after requesting startup.
            wait_timeout_seconds:
                Number of seconds to wait for startup to complete
                before raising a TimeoutError.

        Returns:
            The updated Engine from Firebolt.
        """
        timeout_time = time.time() + wait_timeout_seconds

        engine = self._send_engine_request(ACCOUNT_ENGINE_RESTART_URL)
        logger.info(f"Stopping Engine (engine_id={self.engine_id}, name={self.name})")

        while wait_for_startup and engine.current_status_summary not in {
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED,
        }:
            wait(
                seconds=5,
                timeout_time=timeout_time,
                error_message=(  # noqa: E501
                    f"Could not restart engine within {wait_timeout_seconds} seconds."
                ),
                verbose=False,
            )

            engine = engine.get_latest()

        return engine

    def delete(self) -> Engine:
        """Delete an Engine from Firebolt."""
        response = self._service.client.delete(
            url=ACCOUNT_ENGINE_URL.format(
                account_id=self._service.account_id, engine_id=self.engine_id
            ),
        )
        logger.info(f"Deleting Engine (engine_id={self.engine_id}, name={self.name})")
        return Engine.parse_obj_with_service(
            obj=response.json()["engine"], engine_service=self._service
        )

    def _send_engine_request(self, url: str) -> Engine:
        response = self._service.client.post(
            url=url.format(
                account_id=self._service.account_id, engine_id=self.engine_id
            )
        )
        return Engine.parse_obj_with_service(
            obj=response.json()["engine"], engine_service=self._service
        )


class _EngineCreateRequest(FireboltBaseModel):
    """Helper model for sending Engine create requests."""

    account_id: str
    engine: Engine
    engine_revision: Optional[EngineRevision]



from typing import Optional

from httpx import Timeout

from firebolt.client import Client, log_request, log_response, raise_on_4xx_5xx
from firebolt.client.auth import Token, UsernamePassword
from firebolt.common import Settings
from firebolt.service.provider import get_provider_id

DEFAULT_TIMEOUT_SECONDS: int = 60 * 2


class ResourceManager:
    """
    ResourceManager to access various Firebolt resources:

    - databases
    - engines
    - bindings (the bindings between an engine and a database)
    - engine revisions (versions of an engine)

    Also provides listings of:

    - regions (AWS regions in which engines can run)

    - instance types (AWS instance types which engines can use)
    """

    def __init__(self, settings: Optional[Settings] = None):
        self.settings = settings or Settings()

        auth = self.settings.auth

        # Deprecated: we shouldn't support passing credentials after 1.0 release
        if auth is None:
            if self.settings.access_token:
                auth = Token(self.settings.access_token)
            else:
                auth = UsernamePassword(
                    self.settings.user,
                    self.settings.password.get_secret_value(),
                    self.settings.use_token_cache,
                )

        self.client = Client(
            auth=auth,
            base_url=f"https://{ self.settings.server}",
            account_name=self.settings.account_name,
            api_endpoint=self.settings.server,
            timeout=Timeout(DEFAULT_TIMEOUT_SECONDS),
            event_hooks={
                "request": [log_request],
                "response": [raise_on_4xx_5xx, log_response],
            },
        )
        self.account_id = self.client.account_id
        self._init_services()

    def _init_services(self) -> None:
        # avoid circular import
        from firebolt.service.binding import BindingService
        from firebolt.service.database import DatabaseService
        from firebolt.service.engine import EngineService
        from firebolt.service.engine_revision import EngineRevisionService
        from firebolt.service.instance_type import InstanceTypeService
        from firebolt.service.region import RegionService

        # Cloud Platform Resources (AWS)
        self.regions = RegionService(resource_manager=self)
        self.instance_types = InstanceTypeService(resource_manager=self)
        self.provider_id = get_provider_id(client=self.client)

        # Firebolt Resources
        self.databases = DatabaseService(resource_manager=self)
        self.engines = EngineService(resource_manager=self)
        self.engine_revisions = EngineRevisionService(resource_manager=self)
        self.bindings = BindingService(resource_manager=self)



from __future__ import annotations

import logging
import re
import time
from enum import Enum
from functools import wraps
from inspect import cleandoc
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from aiorwlock import RWLock
from httpx import Response, codes
from pydantic import BaseModel

from firebolt.async_db._types import (
    ColType,
    Column,
    ParameterType,
    RawColType,
    SetParameter,
    parse_type,
    parse_value,
    split_format_sql,
)
from firebolt.async_db.util import is_db_available, is_engine_running
from firebolt.client import AsyncClient
from firebolt.utils.exception import (
    CursorClosedError,
    DataError,
    EngineNotRunningError,
    FireboltDatabaseError,
    OperationalError,
    ProgrammingError,
    QueryNotRunError,
)

if TYPE_CHECKING:
    from firebolt.async_db.connection import Connection

logger = logging.getLogger(__name__)


JSON_OUTPUT_FORMAT = "JSONCompact"


class CursorState(Enum):
    NONE = 1
    ERROR = 2
    DONE = 3
    CLOSED = 4


class Statistics(BaseModel):
    """
    Class for query execution statistics
    """

    elapsed: float
    rows_read: int
    bytes_read: int
    time_before_execution: float
    time_to_execute: float
    scanned_bytes_cache: Optional[float]
    scanned_bytes_storage: Optional[float]


def check_not_closed(func: Callable) -> Callable:
    """(Decorator) ensure cursor is not closed before calling method."""

    @wraps(func)
    def inner(self: Cursor, *args: Any, **kwargs: Any) -> Any:
        if self.closed:
            raise CursorClosedError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


def check_query_executed(func: Callable) -> Callable:
    cleandoc(
        """
        (Decorator) ensure that some query has been executed before
        calling cursor method.
        """
    )

    @wraps(func)
    def inner(self: Cursor, *args: Any, **kwargs: Any) -> Any:
        if self._state == CursorState.NONE:
            raise QueryNotRunError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


class BaseCursor:
    __slots__ = (
        "connection",
        "_arraysize",
        "_client",
        "_state",
        "_descriptions",
        "_statistics",
        "_rowcount",
        "_rows",
        "_idx",
        "_idx_lock",
        "_row_sets",
        "_next_set_idx",
        "_set_parameters",
    )

    default_arraysize = 1

    def __init__(self, client: AsyncClient, connection: Connection):
        self.connection = connection
        self._client = client
        self._arraysize = self.default_arraysize
        # These fields initialized here for type annotations purpose
        self._rows: Optional[List[List[RawColType]]] = None
        self._descriptions: Optional[List[Column]] = None
        self._statistics: Optional[Statistics] = None
        self._row_sets: List[
            Tuple[
                int,
                Optional[List[Column]],
                Optional[Statistics],
                Optional[List[List[RawColType]]],
            ]
        ] = []
        self._set_parameters: Dict[str, Any] = dict()
        self._rowcount = -1
        self._idx = 0
        self._next_set_idx = 0
        self._reset()

    def __del__(self) -> None:
        self.close()

    @property  # type: ignore
    @check_not_closed
    def description(self) -> Optional[List[Column]]:
        """
        Provides information about a single result row of a query

        Attributes:
            * ``name``
            * ``type_code``
            * ``display_size``
            * ``internal_size``
            * ``precision``
            * ``scale``
            * ``null_ok``
        """
        return self._descriptions

    @property  # type: ignore
    @check_not_closed
    def statistics(self) -> Optional[Statistics]:
        """Query execution statistics returned by the backend"""
        return self._statistics

    @property  # type: ignore
    @check_not_closed
    def rowcount(self) -> int:
        """The number of rows produced by last query."""
        return self._rowcount

    @property
    def arraysize(self) -> int:
        """Default number of rows returned by fetchmany."""
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError(
                "Invalid arraysize value type, expected int,"
                f" got {type(value).__name__}"
            )
        self._arraysize = value

    @property
    def closed(self) -> bool:
        """True if connection is closed, False otherwise."""

        return self._state == CursorState.CLOSED

    def close(self) -> None:
        """Terminate an ongoing query (if any) and mark connection as closed."""

        self._state = CursorState.CLOSED
        # remove typecheck skip  after connection is implemented
        self.connection._remove_cursor(self)  # type: ignore

    @check_not_closed
    @check_query_executed
    def nextset(self) -> Optional[bool]:
        """
        Skip to the next available set, discarding any remaining rows
        from the current set.
        Returns True if operation was successful,
        None if there are no more sets to retrive
        """
        return self._pop_next_set()

    def _pop_next_set(self) -> Optional[bool]:
        """
        Same functionality as .nextset, but doesn't check that query has been executed.
        """
        if self._next_set_idx >= len(self._row_sets):
            return None
        (
            self._rowcount,
            self._descriptions,
            self._statistics,
            self._rows,
        ) = self._row_sets[self._next_set_idx]
        self._idx = 0
        self._next_set_idx += 1
        return True

    def flush_parameters(self) -> None:
        self._set_parameters = dict()

    async def _raise_if_error(self, resp: Response) -> None:
        """Raise a proper error if any"""
        if resp.status_code == codes.INTERNAL_SERVER_ERROR:
            raise OperationalError(
                f"Error executing query:\n{resp.read().decode('utf-8')}"
            )
        if resp.status_code == codes.FORBIDDEN:
            if not await is_db_available(self.connection, self.connection.database):
                raise FireboltDatabaseError(
                    f"Database {self.connection.database} does not exist"
                )
            raise ProgrammingError(resp.read().decode("utf-8"))
        if (
            resp.status_code == codes.SERVICE_UNAVAILABLE
            or resp.status_code == codes.NOT_FOUND
        ):
            if not await is_engine_running(self.connection, self.connection.engine_url):
                raise EngineNotRunningError(
                    f"Firebolt engine {self.connection.engine_url} "
                    "needs to be running to run queries against it."
                )
        resp.raise_for_status()

    def _reset(self) -> None:
        """Clear all data stored from previous query."""
        self._state = CursorState.NONE
        self._rows = None
        self._descriptions = None
        self._statistics = None
        self._rowcount = -1
        self._idx = 0
        self._row_sets = []
        self._next_set_idx = 0

    def _row_set_from_response(
        self, response: Response
    ) -> Tuple[
        int,
        Optional[List[Column]],
        Optional[Statistics],
        Optional[List[List[RawColType]]],
    ]:
        """Fetch information about executed query from http response"""

        # Empty response is returned for insert query
        if response.headers.get("content-length", "") == "0":
            return (-1, None, None, None)

        try:
            # Skip parsing floats to properly parse them later
            query_data = response.json(parse_float=str)
            rowcount = int(query_data["rows"])
            descriptions = [
                Column(d["name"], parse_type(d["type"]), None, None, None, None, None)
                for d in query_data["meta"]
            ]
            statistics = Statistics(**query_data["statistics"])
            # Parse data during fetch
            rows = query_data["data"]
            return (rowcount, descriptions, statistics, rows)
        except (KeyError, ValueError) as err:
            raise DataError(f"Invalid query data format: {str(err)}")

    def _append_row_set(
        self,
        row_set: Tuple[
            int,
            Optional[List[Column]],
            Optional[Statistics],
            Optional[List[List[RawColType]]],
        ],
    ) -> None:
        """Store information about executed query."""
        self._row_sets.append(row_set)
        if self._next_set_idx == 0:
            # Populate values for first set
            self._pop_next_set()

    async def _api_request(
        self, query: str, set_parameters: Optional[dict]
    ) -> Response:
        return await self._client.request(
            url="/",
            method="POST",
            params={
                "database": self.connection.database,
                "output_format": JSON_OUTPUT_FORMAT,
                **self._set_parameters,
                **(set_parameters or dict()),
            },
            content=query,
        )

    async def _do_execute(
        self,
        raw_query: str,
        parameters: Sequence[Sequence[ParameterType]],
        set_parameters: Optional[Dict] = None,
        skip_parsing: bool = False,
    ) -> None:
        self._reset()
        if set_parameters is not None:
            logger.warning(
                "Passing set parameters as an argument is deprecated. Please run "
                "a query 'SET <param> = <value>'"
            )
        try:

            if parameters and skip_parsing:
                logger.warning(
                    "Query formatting parameters are provided with skip_parsing."
                    " They will be ignored"
                )

            # Allow users to manually skip parsing for performance improvement
            queries: List[Union[SetParameter, str]] = (
                [raw_query] if skip_parsing else split_format_sql(raw_query, parameters)
            )

            for query in queries:

                start_time = time.time()
                # our CREATE EXTERNAL TABLE queries currently require credentials,
                # so we will skip logging those queries.
                # https://docs.firebolt.io/sql-reference/commands/ddl-commands#create-external-table
                if isinstance(query, SetParameter) or not re.search(
                    "aws_key_id|credentials", query, flags=re.IGNORECASE
                ):
                    logger.debug(f"Running query: {query}")

                # Define type for mypy
                row_set: Tuple[
                    int,
                    Optional[List[Column]],
                    Optional[Statistics],
                    Optional[List[List[RawColType]]],
                ] = (-1, None, None, None)
                if isinstance(query, SetParameter):
                    # Validate parameter by executing simple query with it
                    resp = await self._api_request(
                        "select 1", {query.name: query.value}
                    )
                    # Handle invalid set parameter
                    if resp.status_code == codes.BAD_REQUEST:
                        raise OperationalError(resp.text)
                    await self._raise_if_error(resp)

                    # set parameter passed validation
                    self._set_parameters[query.name] = query.value
                else:
                    resp = await self._api_request(query, set_parameters)
                    await self._raise_if_error(resp)
                    row_set = self._row_set_from_response(resp)

                self._append_row_set(row_set)

                logger.info(
                    f"Query fetched {self.rowcount} rows in"
                    f" {time.time() - start_time} seconds"
                )

            self._state = CursorState.DONE

        except Exception:
            self._state = CursorState.ERROR
            raise

    @check_not_closed
    async def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        set_parameters: Optional[Dict] = None,
        skip_parsing: bool = False,
    ) -> int:
        """Prepare and execute a database query.

        Supported features:
            Parameterized queries: placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.
            Multi-statement queries: multiple statements, provided in a single query
                and separated by semicolon are executed separatelly and sequentially.
                To switch to next statement result, `nextset` method should be used.
            SET statements: to provide additional query execution parameters, execute
                `SET param=value` statement before it. All parameters are stored in
                cursor object until it's closed. They can also be removed with
                `flush_parameters` method call.

        Args:
            query (str): SQL query to execute
            parameters (Optional[Sequence[ParameterType]]): A sequence of substitution
                parameters. Used to replace '?' placeholders inside a query with
                actual values
            set_parameters (Optional[Dict]): List of set parameters to execute
                a query with. DEPRECATED: Use SET SQL statements instead
            skip_parsing (bool): Flag to disable query parsing. This will
                disable parameterized, multi-statement and SET queries,
                while improving performance

        Returns:
            int: Query row count
        """
        params_list = [parameters] if parameters else []
        await self._do_execute(query, params_list, set_parameters, skip_parsing)
        return self.rowcount

    @check_not_closed
    async def executemany(
        self, query: str, parameters_seq: Sequence[Sequence[ParameterType]]
    ) -> int:
        """Prepare and execute a database query.

        Supports providing multiple substitution parameter sets, executing them
        as multiple statements sequentially.

        Supported features:
            Parameterized queries: placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.
            Multi-statement queries: multiple statements, provided in a single query
                and separated by semicolon are executed separatelly and sequentially.
                To switch to next statement result, `nextset` method should be used.
            SET statements: to provide additional query execution parameters, execute
                `SET param=value` statement before it. All parameters are stored in
                cursor object until it's closed. They can also be removed with
                `flush_parameters` method call.

        Args:
            query (str): SQL query to execute
            parameters_seq (Sequence[Sequence[ParameterType]]): A sequence of
               substitution parameter sets. Used to replace '?' placeholders inside a
               query with actual values from each set in a sequence. Resulting queries
               for each subset are executed sequentially.

        Returns:
            int: Query row count
        """
        await self._do_execute(query, parameters_seq)
        return self.rowcount

    def _parse_row(self, row: List[RawColType]) -> List[ColType]:
        """Parse a single data row based on query column types"""
        assert len(row) == len(self.description)
        return [
            parse_value(col, self.description[i].type_code) for i, col in enumerate(row)
        ]

    def _get_next_range(self, size: int) -> Tuple[int, int]:
        cleandoc(
            """
            Return range of next rows of size (if possible),
            and update _idx to point to the end of this range
            """
        )

        if self._rows is None:
            # No elements to take
            raise DataError("no rows to fetch")

        left = self._idx
        right = min(self._idx + size, len(self._rows))
        self._idx = right
        return left, right

    @check_not_closed
    @check_query_executed
    def fetchone(self) -> Optional[List[ColType]]:
        """Fetch the next row of a query result set."""
        left, right = self._get_next_range(1)
        if left == right:
            # We are out of elements
            return None
        assert self._rows is not None
        return self._parse_row(self._rows[left])

    @check_not_closed
    @check_query_executed
    def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        """
        Fetch the next set of rows of a query result,
        cursor.arraysize is default size.
        """
        size = size if size is not None else self.arraysize
        left, right = self._get_next_range(size)
        assert self._rows is not None
        rows = self._rows[left:right]
        return [self._parse_row(row) for row in rows]

    @check_not_closed
    @check_query_executed
    def fetchall(self) -> List[List[ColType]]:
        """Fetch all remaining rows of a query result."""
        left, right = self._get_next_range(self.rowcount)
        assert self._rows is not None
        rows = self._rows[left:right]
        return [self._parse_row(row) for row in rows]

    @check_not_closed
    def setinputsizes(self, sizes: List[int]) -> None:
        """Predefine memory areas for query parameters (does nothing)."""

    @check_not_closed
    def setoutputsize(self, size: int, column: Optional[int] = None) -> None:
        """Set a column buffer size for fetches of large columns (does nothing)."""

    # Context manager support
    @check_not_closed
    def __enter__(self) -> BaseCursor:
        return self

    def __exit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()


class Cursor(BaseCursor):
    """
    Class, responsible for executing asyncio queries to Firebolt Database.
    Should not be created directly,
    use :py:func:`connection.cursor <firebolt.async_db.connection.Connection>`

    Args:
        description: information about a single result row
        rowcount: the number of rows produced by last query
        closed: True if connection is closed, False otherwise
        arraysize: Read/Write, specifies the number of rows to fetch at a time
            with the :py:func:`fetchmany` method

    """

    __slots__ = BaseCursor.__slots__ + ("_async_query_lock",)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._async_query_lock = RWLock()
        super().__init__(*args, **kwargs)

    @wraps(BaseCursor.execute)
    async def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        set_parameters: Optional[Dict] = None,
        skip_parsing: bool = False,
    ) -> int:
        async with self._async_query_lock.writer:
            return await super().execute(
                query, parameters, set_parameters, skip_parsing
            )

    @wraps(BaseCursor.executemany)
    async def executemany(
        self, query: str, parameters_seq: Sequence[Sequence[ParameterType]]
    ) -> int:
        async with self._async_query_lock.writer:
            return await super().executemany(query, parameters_seq)
        """
            Prepare and execute a database query against all parameter
            sequences provided
        """

    @wraps(BaseCursor.fetchone)
    async def fetchone(self) -> Optional[List[ColType]]:
        async with self._async_query_lock.reader:
            return super().fetchone()
        """Fetch the next row of a query result set"""

    @wraps(BaseCursor.fetchmany)
    async def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        async with self._async_query_lock.reader:
            return super().fetchmany(size)
        """fetch the next set of rows of a query result,
          size is cursor.arraysize by default"""

    @wraps(BaseCursor.fetchall)
    async def fetchall(self) -> List[List[ColType]]:
        async with self._async_query_lock.reader:
            return super().fetchall()
        """Fetch all remaining rows of a query result"""

    @wraps(BaseCursor.nextset)
    async def nextset(self) -> None:
        async with self._async_query_lock.reader:
            return super().nextset()

    # Iteration support
    @check_not_closed
    @check_query_executed
    def __aiter__(self) -> Cursor:
        return self

    @check_not_closed
    @check_query_executed
    async def __anext__(self) -> List[ColType]:
        row = await self.fetchone()
        if row is None:
            raise StopAsyncIteration
        return row



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from firebolt.model import FireboltBaseModel


class BindingKey(BaseModel):
    account_id: str
    database_id: str
    engine_id: str


class Binding(FireboltBaseModel):
    """A Binding between an Engine and a Database"""

    binding_key: BindingKey = Field(alias="id")
    is_default_engine: bool = Field(alias="engine_is_default")

    # optional
    current_status: Optional[str]
    health_status: Optional[str]
    create_time: Optional[datetime]
    create_actor: Optional[str]
    last_update_time: Optional[datetime]
    last_update_actor: Optional[str]
    desired_status: Optional[str]

    @property
    def database_id(self) -> str:
        return self.binding_key.database_id

    @property
    def engine_id(self) -> str:
        return self.binding_key.engine_id



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field, PositiveInt

from firebolt.model import FireboltBaseModel
from firebolt.model.instance_type import InstanceTypeKey


class EngineRevisionKey(FireboltBaseModel):
    account_id: str
    engine_id: str
    engine_revision_id: str


class EngineRevisionSpecification(FireboltBaseModel):
    """
    An EngineRevision Specification.

    Notably, it determines which instance types and how many of them its Engine gets.

    See Also: :py:class:`Settings
    <firebolt.model.engine.EngineSettings>`,
    which also contains engine configuration.
    """

    db_compute_instances_type_key: InstanceTypeKey = Field(
        alias="db_compute_instances_type_id"
    )
    db_compute_instances_count: PositiveInt
    db_compute_instances_use_spot: bool = False
    db_version: str = ""
    proxy_instances_type_key: InstanceTypeKey = Field(alias="proxy_instances_type_id")
    proxy_instances_count: PositiveInt = 1
    proxy_version: str = ""


class EngineRevision(FireboltBaseModel):
    """
    A Firebolt Engine revision,
    which contains a Specification (instance types, counts).

    As engines are updated with new settings, revisions are created.
    """

    specification: EngineRevisionSpecification

    # optional
    key: Optional[EngineRevisionKey] = Field(alias="id")
    current_status: Optional[str]
    create_time: Optional[datetime]
    create_actor: Optional[str]
    last_update_time: Optional[datetime]
    last_update_actor: Optional[str]
    desired_status: Optional[str]
    health_status: Optional[str]



from enum import Enum
from types import DynamicClassAttribute


class EngineType(Enum):
    GENERAL_PURPOSE = "GENERAL_PURPOSE"
    DATA_ANALYTICS = "DATA_ANALYTICS"

    @DynamicClassAttribute
    def api_settings_preset_name(self) -> str:
        return {
            EngineType.GENERAL_PURPOSE: "ENGINE_SETTINGS_PRESET_GENERAL_PURPOSE",
            EngineType.DATA_ANALYTICS: "ENGINE_SETTINGS_PRESET_DATA_ANALYTICS",
        }[self]


class WarmupMethod(Enum):
    MINIMAL = "MINIMAL"
    PRELOAD_INDEXES = "PRELOAD_INDEXES"
    PRELOAD_ALL_DATA = "PRELOAD_ALL_DATA"

    @DynamicClassAttribute
    def api_name(self) -> str:
        return {
            WarmupMethod.MINIMAL: "ENGINE_SETTINGS_WARM_UP_MINIMAL",
            WarmupMethod.PRELOAD_INDEXES: "ENGINE_SETTINGS_WARM_UP_INDEXES",
            WarmupMethod.PRELOAD_ALL_DATA: "ENGINE_SETTINGS_WARM_UP_ALL",
        }[self]


class EngineStatus(Enum):
    """
    Detailed engine status.

    See: https://api.dev.firebolt.io/devDocs#operation/coreV1GetEngine
    """

    ENGINE_STATUS_UNSPECIFIED = "ENGINE_STATUS_UNSPECIFIED"
    """ Logical record is created, however underlying infrastructure
    is not initialized.
    In other words this means that engine is stopped."""

    ENGINE_STATUS_CREATED = "ENGINE_STATUS_CREATED"
    """Engine status was created."""

    ENGINE_STATUS_PROVISIONING_PENDING = "ENGINE_STATUS_PROVISIONING_PENDING"
    """ Engine initialization request was sent."""

    ENGINE_STATUS_PROVISIONING_STARTED = "ENGINE_STATUS_PROVISIONING_STARTED"
    """ Engine initialization request was received
    and initialization process started."""

    ENGINE_STATUS_PROVISIONING_FINISHED = "ENGINE_STATUS_PROVISIONING_FINISHED"
    """ Engine initialization was finished successfully."""

    ENGINE_STATUS_PROVISIONING_FAILED = "ENGINE_STATUS_PROVISIONING_FAILED"
    """ Engine initialization failed due to error."""

    ENGINE_STATUS_RUNNING_IDLE = "ENGINE_STATUS_RUNNING_IDLE"
    """ Engine is initialized,
    but there are no running or starting engine revisions."""

    ENGINE_STATUS_RUNNING_REVISION_STARTING = "ENGINE_STATUS_RUNNING_REVISION_STARTING"
    """ Engine is initialized,
    there are no running engine revision but it's starting."""

    ENGINE_STATUS_RUNNING_REVISION_STARTUP_FAILED = (
        "ENGINE_STATUS_RUNNING_REVISION_STARTUP_FAILED"
    )
    """ Engine is initialized,
    initial revision is failed to provision or start."""

    ENGINE_STATUS_RUNNING_REVISION_SERVING = "ENGINE_STATUS_RUNNING_REVISION_SERVING"
    """ Engine is ready (serves an engine revision). """

    ENGINE_STATUS_RUNNING_REVISION_CHANGING = "ENGINE_STATUS_RUNNING_REVISION_CHANGING"
    """ Engine is ready (serves an engine revision),
     zero-downtime replacement revision is starting."""

    ENGINE_STATUS_RUNNING_REVISION_CHANGE_FAILED = (
        "ENGINE_STATUS_RUNNING_REVISION_CHANGE_FAILED"
    )
    """ Engine is ready (serves an engine revision),
     replacement revision failed to provision or start."""

    ENGINE_STATUS_RUNNING_REVISION_RESTARTING = (
        "ENGINE_STATUS_RUNNING_REVISION_RESTARTING"
    )
    """ Engine is initialized,
    replacement of the revision with a downtime is in progress."""

    ENGINE_STATUS_RUNNING_REVISION_RESTART_FAILED = (
        "ENGINE_STATUS_RUNNING_REVISION_RESTART_FAILED"
    )
    """ Engine is initialized,
    replacement revision failed to provision or start."""

    ENGINE_STATUS_RUNNING_REVISIONS_TERMINATING = (
        "ENGINE_STATUS_RUNNING_REVISIONS_TERMINATING"
    )
    """ Engine is initialized,
    all child revisions are being terminated."""

    # Engine termination request was sent.
    ENGINE_STATUS_TERMINATION_PENDING = "ENGINE_STATUS_TERMINATION_PENDING"
    """ Engine termination request was sent."""

    ENGINE_STATUS_TERMINATION_ST = "ENGINE_STATUS_TERMINATION_STARTED"
    """ Engine termination started."""

    ENGINE_STATUS_TERMINATION_FIN = "ENGINE_STATUS_TERMINATION_FINISHED"
    """ Engine termination finished."""

    ENGINE_STATUS_TERMINATION_F = "ENGINE_STATUS_TERMINATION_FAILED"
    """ Engine termination failed."""

    ENGINE_STATUS_DELETED = "ENGINE_STATUS_DELETED"
    """ Engine is soft-deleted."""


class EngineStatusSummary(Enum):
    """
    Engine summary status.

    See: https://api.dev.firebolt.io/devDocs#operation/coreV1GetEngine
    """

    ENGINE_STATUS_SUMMARY_UNSPECIFIED = "ENGINE_STATUS_SUMMARY_UNSPECIFIED"
    """Status unspecified"""

    ENGINE_STATUS_SUMMARY_STOPPED = "ENGINE_STATUS_SUMMARY_STOPPED"
    """ Fully stopped."""

    ENGINE_STATUS_SUMMARY_STARTING = "ENGINE_STATUS_SUMMARY_STARTING"
    """ Provisioning process is in progress.
     We are creating cloud infra for this engine."""

    ENGINE_STATUS_SUMMARY_STARTING_INITIALIZING = (
        "ENGINE_STATUS_SUMMARY_STARTING_INITIALIZING"
    )
    """ Provisioning process is complete.
     We are now waiting for PackDB cluster to initialize and start."""

    ENGINE_STATUS_SUMMARY_RUNNING = "ENGINE_STATUS_SUMMARY_RUNNING"
    """ Fully started.
     Engine is ready to serve requests."""

    ENGINE_STATUS_SUMMARY_UPGRADING = "ENGINE_STATUS_SUMMARY_UPGRADING"
    """ Version of the PackDB is changing.
     This is zero downtime operation that does not affect engine work.
     This status is reserved for future use (not used fow now)."""

    ENGINE_STATUS_SUMMARY_RESTARTING = "ENGINE_STATUS_SUMMARY_RESTARTING"
    """ Hard restart (full stop/start cycle) is in progress.
    Underlying infrastructure is being recreated."""

    ENGINE_STATUS_SUMMARY_RESTARTING_INITIALIZING = (
        "ENGINE_STATUS_SUMMARY_RESTARTING_INITIALIZING"
    )
    """ Hard restart (full stop/start cycle) is in progress.
     Underlying infrastructure is ready, waiting for
     PackDB cluster to initialize and start.
     This status is logically the same as ENGINE_STATUS_SUMMARY_STARTING_INITIALIZING,
     but used during restart cycle."""

    ENGINE_STATUS_SUMMARY_REPAIRING = "ENGINE_STATUS_SUMMARY_REPAIRING"
    """ Underlying infrastructure has issues and is being repaired.
     Engine is still running, but it's not fully healthy and some queries may fail."""

    ENGINE_STATUS_SUMMARY_STOPPING = "ENGINE_STATUS_SUMMARY_STOPPING"
    """ Stop is in progress."""

    ENGINE_STATUS_SUMMARY_DELETING = "ENGINE_STATUS_SUMMARY_DELETING"
    """ Termination is in progress.
     All infrastructure that belongs to this engine will be completely destroyed."""

    ENGINE_STATUS_SUMMARY_DELETED = "ENGINE_STATUS_SUMMARY_DELETED"
    """ Infrastructure is terminated, engine data is deleted."""

    ENGINE_STATUS_SUMMARY_FAILED = "ENGINE_STATUS_SUMMARY_FAILED"
    """ Failed to start or stop.
     This status only indicates that there were issues during provisioning operations.
     If engine enters this status,
     all infrastructure should be stopped/terminated already."""


class EngineOrder(Enum):
    ENGINE_ORDER_UNSPECIFIED = "ENGINE_ORDER_UNSPECIFIED"
    ENGINE_ORDER_NAME_ASC = "ENGINE_ORDER_NAME_ASC"
    ENGINE_ORDER_NAME_DESC = "ENGINE_ORDER_NAME_DESC"
    ENGINE_ORDER_COMPUTE_REGION_ID_ASC = "ENGINE_ORDER_COMPUTE_REGION_ID_ASC"
    ENGINE_ORDER_COMPUTE_REGION_ID_DESC = "ENGINE_ORDER_COMPUTE_REGION_ID_DESC"
    ENGINE_ORDER_CURRENT_STATUS_ASC = "ENGINE_ORDER_CURRENT_STATUS_ASC"
    ENGINE_ORDER_CURRENT_STATUS_DESC = "ENGINE_ORDER_CURRENT_STATUS_DESC"
    ENGINE_ORDER_CREATE_TIME_ASC = "ENGINE_ORDER_CREATE_TIME_ASC"
    ENGINE_ORDER_CREATE_TIME_DESC = "ENGINE_ORDER_CREATE_TIME_DESC"
    ENGINE_ORDER_CREATE_ACTOR_ASC = "ENGINE_ORDER_CREATE_ACTOR_ASC"
    ENGINE_ORDER_CREATE_ACTOR_DESC = "ENGINE_ORDER_CREATE_ACTOR_DESC"
    ENGINE_ORDER_LAST_UPDATE_TIME_ASC = "ENGINE_ORDER_LAST_UPDATE_TIME_ASC"
    ENGINE_ORDER_LAST_UPDATE_TIME_DESC = "ENGINE_ORDER_LAST_UPDATE_TIME_DESC"
    ENGINE_ORDER_LAST_UPDATE_ACTOR_ASC = "ENGINE_ORDER_LAST_UPDATE_ACTOR_ASC"
    ENGINE_ORDER_LAST_UPDATE_ACTOR_DESC = "ENGINE_ORDER_LAST_UPDATE_ACTOR_DESC"
    ENGINE_ORDER_LATEST_REVISION_CURRENT_STATUS_ASC = (
        "ENGINE_ORDER_LATEST_REVISION_CURRENT_STATUS_ASC"
    )
    ENGINE_ORDER_LATEST_REVISION_CURRENT_STATUS_DESC = (
        "ENGINE_ORDER_LATEST_REVISION_CURRENT_STATUS_DESC"
    )
    ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_COUNT_ASC = (
        "ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_COUNT_ASC"
    )
    ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_COUNT_DESC = (
        "ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_COUNT_DESC"
    )
    ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_TYPE_ID_ASC = (
        "ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_TYPE_ID_ASC"
    )
    ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_TYPE_ID_DESC = (
        "ENGINE_ORDER_LATEST_REVISION_SPECIFICATION_DB_COMPUTE_INSTANCES_TYPE_ID_DESC"
    )


class DatabaseOrder(Enum):
    DATABASE_ORDER_UNSPECIFIED = "DATABASE_ORDER_UNSPECIFIED"
    DATABASE_ORDER_NAME_ASC = "DATABASE_ORDER_NAME_ASC"
    DATABASE_ORDER_NAME_DESC = "DATABASE_ORDER_NAME_DESC"
    DATABASE_ORDER_COMPUTE_REGION_ID_ASC = "DATABASE_ORDER_COMPUTE_REGION_ID_ASC"
    DATABASE_ORDER_COMPUTE_REGION_ID_DESC = "DATABASE_ORDER_COMPUTE_REGION_ID_DESC"
    DATABASE_ORDER_DATA_SIZE_FULL_ASC = "DATABASE_ORDER_DATA_SIZE_FULL_ASC"
    DATABASE_ORDER_DATA_SIZE_FULL_DESC = "DATABASE_ORDER_DATA_SIZE_FULL_DESC"
    DATABASE_ORDER_DATA_SIZE_COMPRESSED_ASC = "DATABASE_ORDER_DATA_SIZE_COMPRESSED_ASC"
    DATABASE_ORDER_DATA_SIZE_COMPRESSED_DESC = (
        "DATABASE_ORDER_DATA_SIZE_COMPRESSED_DESC"
    )
    DATABASE_ORDER_CREATE_TIME_ASC = "DATABASE_ORDER_CREATE_TIME_ASC"
    DATABASE_ORDER_CREATE_TIME_DESC = "DATABASE_ORDER_CREATE_TIME_DESC"
    DATABASE_ORDER_CREATE_ACTOR_ASC = "DATABASE_ORDER_CREATE_ACTOR_ASC"
    DATABASE_ORDER_CREATE_ACTOR_DESC = "DATABASE_ORDER_CREATE_ACTOR_DESC"
    DATABASE_ORDER_LAST_UPDATE_TIME_ASC = "DATABASE_ORDER_LAST_UPDATE_TIME_ASC"
    DATABASE_ORDER_LAST_UPDATE_TIME_DESC = "DATABASE_ORDER_LAST_UPDATE_TIME_DESC"
    DATABASE_ORDER_LAST_UPDATE_ACTOR_ASC = "DATABASE_ORDER_LAST_UPDATE_ACTOR_ASC"
    DATABASE_ORDER_LAST_UPDATE_ACTOR_DESC = "DATABASE_ORDER_LAST_UPDATE_ACTOR_DESC"



from firebolt.model.engine_revision import EngineRevision, EngineRevisionKey
from firebolt.service.base import BaseService
from firebolt.utils.urls import ACCOUNT_ENGINE_REVISION_URL


class EngineRevisionService(BaseService):
    def get_by_id(self, engine_id: str, engine_revision_id: str) -> EngineRevision:
        """
        Get an EngineRevision from Firebolt by engine_id and engine_revision_id.
        """

        return self.get_by_key(
            EngineRevisionKey(
                account_id=self.account_id,
                engine_id=engine_id,
                engine_revision_id=engine_revision_id,
            )
        )

    def get_by_key(self, key: EngineRevisionKey) -> EngineRevision:
        """
        Fetch an EngineRevision from Firebolt by its key.

        Args:
            key: Key of the desired EngineRevision.

        Returns:
            The requested EngineRevision
        """

        response = self.client.get(
            url=ACCOUNT_ENGINE_REVISION_URL.format(
                account_id=key.account_id,
                engine_id=key.engine_id,
                revision_id=key.engine_revision_id,
            ),
        )
        engine_spec: dict = response.json()["engine_revision"]
        return EngineRevision.parse_obj(engine_spec)



from datetime import datetime
from typing import Optional

from pydantic import Field

from firebolt.model import FireboltBaseModel


class InstanceTypeKey(FireboltBaseModel, frozen=True):  # type: ignore
    provider_id: str
    region_id: str
    instance_type_id: str


class InstanceType(FireboltBaseModel):
    key: InstanceTypeKey = Field(alias="id")
    name: str

    # optional
    is_spot_available: Optional[bool]
    cpu_virtual_cores_count: Optional[int]
    memory_size_bytes: Optional[int]
    storage_size_bytes: Optional[int]
    price_per_hour_cents: Optional[float]
    create_time: Optional[datetime]
    last_update_time: Optional[datetime]
